// This file was automatically generated by bifcl from /usr/obj/ports/zeek-8.2.0/zeek-8.2.0/src/input/input.bif (alternative mode).

#pragma once

#if defined(ZEEK_IN_NETVAR) || ! defined(_usr_obj_ports_zeek_8_2_0_build_aarch64_src_input_input_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _usr_obj_ports_zeek_8_2_0_build_aarch64_src_input_input_bif
#define _usr_obj_ports_zeek_8_2_0_build_aarch64_src_input_input_bif
#include "zeek/zeek-bif.h"
#endif
#endif

// NOLINTNEXTLINE(performance-enum-size)
namespace zeek::BifEnum::Input { enum Event {
	EVENT_NEW = 0,
	EVENT_CHANGED = 1,
	EVENT_REMOVED = 2,
}; }
namespace zeek::BifType::Enum::Input { extern zeek::IntrusivePtr<zeek::EnumType> Event; }
namespace zeek::BifType::Record::Input { extern zeek::IntrusivePtr<zeek::RecordType> TableDescription; }
namespace zeek::BifType::Record::Input { extern zeek::IntrusivePtr<zeek::RecordType> EventDescription; }
namespace zeek::BifType::Record::Input { extern zeek::IntrusivePtr<zeek::RecordType> AnalysisDescription; }
namespace zeek::BifFunc::Input { extern zeek::ValPtr __create_table_stream_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Input { extern zeek::ValPtr __create_event_stream_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Input { extern zeek::ValPtr __create_analysis_stream_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Input { extern zeek::ValPtr __remove_stream_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Input { extern zeek::ValPtr __force_update_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifConst::Input { extern int accept_unsupported_types; }

#endif
