// This file was automatically generated by bifcl from /usr/obj/ports/zeek-8.2.0/zeek-8.2.0/src/broker/messaging.bif (alternative mode).

#pragma once

#if defined(ZEEK_IN_NETVAR) || ! defined(_usr_obj_ports_zeek_8_2_0_build_aarch64_src_broker_messaging_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _usr_obj_ports_zeek_8_2_0_build_aarch64_src_broker_messaging_bif
#define _usr_obj_ports_zeek_8_2_0_build_aarch64_src_broker_messaging_bif
#include "zeek/zeek-bif.h"
#endif
#endif

namespace zeek::BifType::Record::Broker { extern zeek::IntrusivePtr<zeek::RecordType> Event; }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr make_event_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr publish_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __flush_logs_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __publish_id_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __subscribe_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __forward_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __unsubscribe_bif(zeek::detail::Frame* frame, const zeek::Args*); }

#endif
