// This file was automatically generated by bifcl from /usr/obj/ports/zeek-8.2.0/zeek-8.2.0/src/analyzer/protocol/dce-rpc/consts.bif (plugin mode).

#pragma once

#if defined(ZEEK_IN_NETVAR) || ! defined(_usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_dce_rpc_consts_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_dce_rpc_consts_bif
#define _usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_dce_rpc_consts_bif
#include "zeek/zeek-bif.h"
#endif
#endif

namespace zeek::BifConst::DCE_RPC { extern zeek_uint_t max_cmd_reassembly; }
namespace zeek::BifConst::DCE_RPC { extern zeek_uint_t max_frag_data; }

#endif
