// See the file "COPYING" in the main distribution directory for copyright.

#pragma once

namespace zeek::spicy::configuration {

int ZeekVersionNumber();
const char* ZeekVersionString();

} // namespace zeek::spicy::configuration
