// This file was automatically generated by bifcl from /usr/obj/ports/zeek-8.2.0/zeek-8.2.0/src/storage/storage-events.bif (alternative mode).

#pragma once

#if defined(ZEEK_IN_NETVAR) || ! defined(_usr_obj_ports_zeek_8_2_0_build_aarch64_src_storage_storage_events_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _usr_obj_ports_zeek_8_2_0_build_aarch64_src_storage_storage_events_bif
#define _usr_obj_ports_zeek_8_2_0_build_aarch64_src_storage_storage_events_bif
#include "zeek/zeek-bif.h"
#endif
#endif

namespace Storage { extern zeek::EventHandlerPtr backend_opened;  }
namespace zeek::BifEvent::Storage { void enqueue_backend_opened(zeek::analyzer::Analyzer* analyzer, zeek::IntrusivePtr<zeek::Val> tag, zeek::IntrusivePtr<zeek::Val> options);  }
namespace Storage { extern zeek::EventHandlerPtr backend_lost;  }
namespace zeek::BifEvent::Storage { void enqueue_backend_lost(zeek::analyzer::Analyzer* analyzer, zeek::IntrusivePtr<zeek::Val> tag, zeek::IntrusivePtr<zeek::Val> options, zeek::IntrusivePtr<zeek::StringVal> reason);  }

#endif
