// This file was automatically generated by bifcl from /usr/obj/ports/zeek-8.2.0/zeek-8.2.0/src/file_analysis/analyzer/x509/functions.bif (plugin mode).

#pragma once

#if defined(ZEEK_IN_NETVAR) || ! defined(_usr_obj_ports_zeek_8_2_0_build_aarch64_src_file_analysis_analyzer_x509_functions_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _usr_obj_ports_zeek_8_2_0_build_aarch64_src_file_analysis_analyzer_x509_functions_bif
#define _usr_obj_ports_zeek_8_2_0_build_aarch64_src_file_analysis_analyzer_x509_functions_bif
#include "zeek/zeek-bif.h"
#endif
#endif

namespace zeek::BifFunc { extern zeek::ValPtr x509_parse_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc { extern zeek::ValPtr x509_from_der_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc { extern zeek::ValPtr x509_get_certificate_string_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc { extern zeek::ValPtr x509_ocsp_verify_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc { extern zeek::ValPtr x509_verify_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc { extern zeek::ValPtr sct_verify_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc { extern zeek::ValPtr x509_subject_name_hash_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc { extern zeek::ValPtr x509_issuer_name_hash_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc { extern zeek::ValPtr x509_spki_hash_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc { extern zeek::ValPtr x509_set_certificate_cache_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc { extern zeek::ValPtr x509_set_certificate_cache_hit_callback_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc { extern zeek::ValPtr x509_check_hostname_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc { extern zeek::ValPtr x509_check_cert_hostname_bif(zeek::detail::Frame* frame, const zeek::Args*); }

#endif
