// This file was automatically generated by bifcl from /usr/obj/ports/zeek-8.2.0/zeek-8.2.0/src/logging/logging.bif (alternative mode).

#pragma once

#if defined(ZEEK_IN_NETVAR) || ! defined(_usr_obj_ports_zeek_8_2_0_build_aarch64_src_logging_logging_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _usr_obj_ports_zeek_8_2_0_build_aarch64_src_logging_logging_bif
#define _usr_obj_ports_zeek_8_2_0_build_aarch64_src_logging_logging_bif
#include "zeek/zeek-bif.h"
#endif
#endif

namespace zeek::BifType::Record::Log { extern zeek::IntrusivePtr<zeek::RecordType> Filter; }
namespace zeek::BifType::Record::Log { extern zeek::IntrusivePtr<zeek::RecordType> Stream; }
namespace zeek::BifType::Record::Log { extern zeek::IntrusivePtr<zeek::RecordType> RotationInfo; }
namespace zeek::BifType::Record::Log { extern zeek::IntrusivePtr<zeek::RecordType> RotationFmtInfo; }
// NOLINTNEXTLINE(performance-enum-size)
namespace zeek::BifEnum::Log { enum PrintLogType {
	REDIRECT_NONE,
	REDIRECT_STDOUT,
	REDIRECT_ALL,
}; }
namespace zeek::BifType::Enum::Log { extern zeek::IntrusivePtr<zeek::EnumType> PrintLogType; }
namespace zeek::BifFunc::Log { extern zeek::ValPtr __create_stream_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Log { extern zeek::ValPtr __remove_stream_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Log { extern zeek::ValPtr __enable_stream_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Log { extern zeek::ValPtr __disable_stream_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Log { extern zeek::ValPtr __add_filter_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Log { extern zeek::ValPtr __remove_filter_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Log { extern zeek::ValPtr __write_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Log { extern zeek::ValPtr __set_buf_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Log { extern zeek::ValPtr __flush_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Log { extern zeek::ValPtr flush_all_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Log { extern zeek::ValPtr __delay_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Log { extern zeek::ValPtr __delay_finish_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Log { extern zeek::ValPtr __set_max_delay_interval_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Log { extern zeek::ValPtr __set_max_delay_queue_size_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Log { extern zeek::ValPtr __get_delay_queue_size_bif(zeek::detail::Frame* frame, const zeek::Args*); }

#endif
