// This file was automatically generated by bifcl from /usr/obj/ports/zeek-8.2.0/zeek-8.2.0/src/analyzer/protocol/smb/smb1_com_transaction2.bif (plugin mode).

#pragma once

#if defined(ZEEK_IN_NETVAR) || ! defined(_usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_smb_smb1_com_transaction2_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_smb_smb1_com_transaction2_bif
#define _usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_smb_smb1_com_transaction2_bif
#include "zeek/zeek-bif.h"
#endif
#endif

extern zeek::EventHandlerPtr smb1_transaction2_request; 
namespace zeek::BifEvent { void enqueue_smb1_transaction2_request(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> hdr, zeek::IntrusivePtr<zeek::Val> args, zeek_uint_t sub_cmd);  }
extern zeek::EventHandlerPtr smb1_trans2_find_first2_request; 
namespace zeek::BifEvent { void enqueue_smb1_trans2_find_first2_request(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> hdr, zeek::IntrusivePtr<zeek::Val> args);  }
extern zeek::EventHandlerPtr smb1_trans2_query_path_info_request; 
namespace zeek::BifEvent { void enqueue_smb1_trans2_query_path_info_request(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> hdr, zeek::IntrusivePtr<zeek::StringVal> file_name);  }
extern zeek::EventHandlerPtr smb1_trans2_get_dfs_referral_request; 
namespace zeek::BifEvent { void enqueue_smb1_trans2_get_dfs_referral_request(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> hdr, zeek::IntrusivePtr<zeek::StringVal> file_name);  }
namespace zeek::BifType::Record::SMB1 { extern zeek::IntrusivePtr<zeek::RecordType> Find_First2_Request_Args; }
namespace zeek::BifType::Record::SMB1 { extern zeek::IntrusivePtr<zeek::RecordType> Find_First2_Response_Args; }
namespace zeek::BifType::Record::SMB1 { extern zeek::IntrusivePtr<zeek::RecordType> Trans2_Args; }

#endif
