// This file was automatically generated by bifcl from /usr/obj/ports/zeek-8.2.0/zeek-8.2.0/src/file_analysis/file_analysis.bif (alternative mode).

#pragma once

#if defined(ZEEK_IN_NETVAR) || ! defined(_usr_obj_ports_zeek_8_2_0_build_aarch64_src_file_analysis_file_analysis_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _usr_obj_ports_zeek_8_2_0_build_aarch64_src_file_analysis_file_analysis_bif
#define _usr_obj_ports_zeek_8_2_0_build_aarch64_src_file_analysis_file_analysis_bif
#include "zeek/zeek-bif.h"
#endif
#endif

namespace zeek::BifType::Record::Files { extern zeek::IntrusivePtr<zeek::RecordType> AnalyzerArgs; }
namespace zeek::BifFunc::Files { extern zeek::ValPtr __set_timeout_interval_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Files { extern zeek::ValPtr __enable_reassembly_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Files { extern zeek::ValPtr __disable_reassembly_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Files { extern zeek::ValPtr __set_reassembly_buffer_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Files { extern zeek::ValPtr __enable_analyzer_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Files { extern zeek::ValPtr __disable_analyzer_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Files { extern zeek::ValPtr __analyzer_enabled_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Files { extern zeek::ValPtr __add_analyzer_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Files { extern zeek::ValPtr __remove_analyzer_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Files { extern zeek::ValPtr __stop_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Files { extern zeek::ValPtr __analyzer_name_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Files { extern zeek::ValPtr __file_exists_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Files { extern zeek::ValPtr __lookup_file_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc { extern zeek::ValPtr set_file_handle_bif(zeek::detail::Frame* frame, const zeek::Args*); }

#endif
