// This file was automatically generated by bifcl from /usr/obj/ports/zeek-8.2.0/zeek-8.2.0/src/analyzer/protocol/syslog/legacy/events.bif (plugin mode).

#pragma once

#if defined(ZEEK_IN_NETVAR) || ! defined(_usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_syslog_legacy_events_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_syslog_legacy_events_bif
#define _usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_syslog_legacy_events_bif
#include "zeek/zeek-bif.h"
#endif
#endif

extern zeek::EventHandlerPtr syslog_message; 
namespace zeek::BifEvent { void enqueue_syslog_message(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek_uint_t facility, zeek_uint_t severity, zeek::IntrusivePtr<zeek::StringVal> msg);  }

#endif
