// This file was automatically generated by bifcl from /usr/obj/ports/zeek-8.2.0/zeek-8.2.0/src/analyzer/protocol/smb/events.bif (plugin mode).

#pragma once

#if defined(ZEEK_IN_NETVAR) || ! defined(_usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_smb_events_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_smb_events_bif
#define _usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_smb_events_bif
#include "zeek/zeek-bif.h"
#endif
#endif

extern zeek::EventHandlerPtr smb_pipe_connect_heuristic; 
namespace zeek::BifEvent { void enqueue_smb_pipe_connect_heuristic(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c);  }
extern zeek::EventHandlerPtr smb_discarded_dce_rpc_analyzers; 
namespace zeek::BifEvent { void enqueue_smb_discarded_dce_rpc_analyzers(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c);  }

#endif
