// This file was automatically generated by bifcl from /usr/obj/ports/zeek-8.2.0/zeek-8.2.0/src/analyzer/protocol/smb/smb2_com_tree_disconnect.bif (plugin mode).

#pragma once

#if defined(ZEEK_IN_NETVAR) || ! defined(_usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_smb_smb2_com_tree_disconnect_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_smb_smb2_com_tree_disconnect_bif
#define _usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_smb_smb2_com_tree_disconnect_bif
#include "zeek/zeek-bif.h"
#endif
#endif

extern zeek::EventHandlerPtr smb2_tree_disconnect_request; 
namespace zeek::BifEvent { void enqueue_smb2_tree_disconnect_request(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> hdr);  }
extern zeek::EventHandlerPtr smb2_tree_disconnect_response; 
namespace zeek::BifEvent { void enqueue_smb2_tree_disconnect_response(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> hdr);  }

#endif
