// This file was automatically generated by bifcl from /usr/obj/ports/zeek-8.2.0/zeek-8.2.0/src/iosource/pcap/pcap.bif (alternative mode).

#pragma once

#if defined(ZEEK_IN_NETVAR) || ! defined(_usr_obj_ports_zeek_8_2_0_build_aarch64_src_iosource_pcap_pcap_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _usr_obj_ports_zeek_8_2_0_build_aarch64_src_iosource_pcap_pcap_bif
#define _usr_obj_ports_zeek_8_2_0_build_aarch64_src_iosource_pcap_pcap_bif
#include "zeek/zeek-bif.h"
#endif
#endif

namespace zeek::BifConst::Pcap { extern zeek_uint_t snaplen; }
namespace zeek::BifConst::Pcap { extern zeek_uint_t bufsize; }
namespace zeek::BifConst::Pcap { extern zeek_uint_t bufsize_offline_bytes; }
namespace zeek::BifConst::Pcap { extern double non_fd_timeout; }
namespace zeek::BifFunc::Pcap { extern zeek::ValPtr precompile_pcap_filter_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Pcap { extern zeek::ValPtr install_pcap_filter_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Pcap { extern zeek::ValPtr error_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Pcap { extern zeek::ValPtr get_filter_state_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Pcap { extern zeek::ValPtr get_filter_state_string_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Pcap { extern zeek::ValPtr findalldevs_bif(zeek::detail::Frame* frame, const zeek::Args*); }

#endif
