// This file was automatically generated by bifcl from /usr/obj/ports/zeek-8.2.0/zeek-8.2.0/src/analyzer/protocol/pop3/events.bif (plugin mode).

#pragma once

#if defined(ZEEK_IN_NETVAR) || ! defined(_usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_pop3_events_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_pop3_events_bif
#define _usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_pop3_events_bif
#include "zeek/zeek-bif.h"
#endif
#endif

extern zeek::EventHandlerPtr pop3_request; 
namespace zeek::BifEvent { void enqueue_pop3_request(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, int is_orig, zeek::IntrusivePtr<zeek::StringVal> command, zeek::IntrusivePtr<zeek::StringVal> arg);  }
extern zeek::EventHandlerPtr pop3_reply; 
namespace zeek::BifEvent { void enqueue_pop3_reply(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, int is_orig, zeek::IntrusivePtr<zeek::StringVal> cmd, zeek::IntrusivePtr<zeek::StringVal> msg);  }
extern zeek::EventHandlerPtr pop3_data; 
namespace zeek::BifEvent { void enqueue_pop3_data(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, int is_orig, zeek::IntrusivePtr<zeek::StringVal> data);  }
extern zeek::EventHandlerPtr pop3_unexpected; 
namespace zeek::BifEvent { void enqueue_pop3_unexpected(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, int is_orig, zeek::IntrusivePtr<zeek::StringVal> msg, zeek::IntrusivePtr<zeek::StringVal> detail);  }
extern zeek::EventHandlerPtr pop3_starttls; 
namespace zeek::BifEvent { void enqueue_pop3_starttls(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c);  }
extern zeek::EventHandlerPtr pop3_login_success; 
namespace zeek::BifEvent { void enqueue_pop3_login_success(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, int is_orig, zeek::IntrusivePtr<zeek::StringVal> user, zeek::IntrusivePtr<zeek::StringVal> password);  }
extern zeek::EventHandlerPtr pop3_login_failure; 
namespace zeek::BifEvent { void enqueue_pop3_login_failure(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, int is_orig, zeek::IntrusivePtr<zeek::StringVal> user, zeek::IntrusivePtr<zeek::StringVal> password);  }

#endif
