#
# Settings that specify the version of ZeekControl
#

VERSION = "2.6.0-56"
ZEEKBASE = "/usr/local"
CFGFILE = "/etc/zeek/zeekctl.cfg"
ZEEKSCRIPTDIR = "/usr/local/share/zeek"
LIBDIR = "/usr/local/lib"
LIBDIRINTERNAL = "/usr/local/lib/zeek/python"
