// This file was automatically generated by bifcl from /usr/obj/ports/zeek-8.2.0/zeek-8.2.0/src/input/readers/ascii/ascii.bif (plugin mode).

#pragma once

#if defined(ZEEK_IN_NETVAR) || ! defined(_usr_obj_ports_zeek_8_2_0_build_aarch64_src_input_readers_ascii_ascii_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _usr_obj_ports_zeek_8_2_0_build_aarch64_src_input_readers_ascii_ascii_bif
#define _usr_obj_ports_zeek_8_2_0_build_aarch64_src_input_readers_ascii_ascii_bif
#include "zeek/zeek-bif.h"
#endif
#endif

namespace zeek::BifConst::InputAscii { extern zeek::IntrusivePtr<zeek::StringVal> separator; }
namespace zeek::BifConst::InputAscii { extern zeek::IntrusivePtr<zeek::StringVal> set_separator; }
namespace zeek::BifConst::InputAscii { extern zeek::IntrusivePtr<zeek::StringVal> empty_field; }
namespace zeek::BifConst::InputAscii { extern zeek::IntrusivePtr<zeek::StringVal> unset_field; }
namespace zeek::BifConst::InputAscii { extern int fail_on_invalid_lines; }
namespace zeek::BifConst::InputAscii { extern int fail_on_file_problem; }
namespace zeek::BifConst::InputAscii { extern zeek::IntrusivePtr<zeek::StringVal> path_prefix; }

#endif
