// This file was automatically generated by bifcl from /usr/obj/ports/zeek-8.2.0/zeek-8.2.0/src/analyzer/protocol/krb/types.bif (plugin mode).

#pragma once

#if defined(ZEEK_IN_NETVAR) || ! defined(_usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_krb_types_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_krb_types_bif
#define _usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_krb_types_bif
#include "zeek/zeek-bif.h"
#endif
#endif

namespace zeek::BifConst::KRB { extern zeek::IntrusivePtr<zeek::StringVal> keytab; }
namespace zeek::BifType::Record::KRB { extern zeek::IntrusivePtr<zeek::RecordType> Error_Msg; }
namespace zeek::BifType::Record::KRB { extern zeek::IntrusivePtr<zeek::RecordType> SAFE_Msg; }
namespace zeek::BifType::Record::KRB { extern zeek::IntrusivePtr<zeek::RecordType> KDC_Options; }
namespace zeek::BifType::Record::KRB { extern zeek::IntrusivePtr<zeek::RecordType> AP_Options; }
namespace zeek::BifType::Record::KRB { extern zeek::IntrusivePtr<zeek::RecordType> Type_Value; }
namespace zeek::BifType::Record::KRB { extern zeek::IntrusivePtr<zeek::RecordType> Encrypted_Data; }
namespace zeek::BifType::Record::KRB { extern zeek::IntrusivePtr<zeek::RecordType> Ticket; }
namespace zeek::BifType::Vector::KRB { extern zeek::IntrusivePtr<zeek::VectorType> Ticket_Vector; }
namespace zeek::BifType::Record::KRB { extern zeek::IntrusivePtr<zeek::RecordType> Host_Address; }
namespace zeek::BifType::Record::KRB { extern zeek::IntrusivePtr<zeek::RecordType> KDC_Request; }
namespace zeek::BifType::Record::KRB { extern zeek::IntrusivePtr<zeek::RecordType> KDC_Response; }

#endif
