// This file was automatically generated by bifcl from /usr/obj/ports/zeek-8.2.0/zeek-8.2.0/src/logging/writers/ascii/ascii.bif (plugin mode).

#pragma once

#if defined(ZEEK_IN_NETVAR) || ! defined(_usr_obj_ports_zeek_8_2_0_build_aarch64_src_logging_writers_ascii_ascii_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _usr_obj_ports_zeek_8_2_0_build_aarch64_src_logging_writers_ascii_ascii_bif
#define _usr_obj_ports_zeek_8_2_0_build_aarch64_src_logging_writers_ascii_ascii_bif
#include "zeek/zeek-bif.h"
#endif
#endif

namespace zeek::BifConst::LogAscii { extern int output_to_stdout; }
namespace zeek::BifConst::LogAscii { extern int include_meta; }
namespace zeek::BifConst::LogAscii { extern zeek::IntrusivePtr<zeek::StringVal> meta_prefix; }
namespace zeek::BifConst::LogAscii { extern zeek::IntrusivePtr<zeek::StringVal> separator; }
namespace zeek::BifConst::LogAscii { extern zeek::IntrusivePtr<zeek::StringVal> set_separator; }
namespace zeek::BifConst::LogAscii { extern zeek::IntrusivePtr<zeek::StringVal> empty_field; }
namespace zeek::BifConst::LogAscii { extern zeek::IntrusivePtr<zeek::StringVal> unset_field; }
namespace zeek::BifConst::LogAscii { extern int use_json; }
namespace zeek::BifConst::LogAscii { extern int enable_leftover_log_rotation; }
namespace zeek::BifConst::LogAscii { extern int enable_utf_8; }
namespace zeek::BifConst::LogAscii { extern zeek::IntrusivePtr<zeek::Val> json_timestamps; }
namespace zeek::BifConst::LogAscii { extern int json_include_unset_fields; }
namespace zeek::BifConst::LogAscii { extern zeek_uint_t gzip_level; }
namespace zeek::BifConst::LogAscii { extern zeek::IntrusivePtr<zeek::StringVal> gzip_file_extension; }

#endif
