// This file was automatically generated by bifcl from /usr/obj/ports/zeek-8.2.0/zeek-8.2.0/src/cluster/websocket/events.bif (plugin mode).

#pragma once

#if defined(ZEEK_IN_NETVAR) || ! defined(_usr_obj_ports_zeek_8_2_0_build_aarch64_src_cluster_websocket_events_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _usr_obj_ports_zeek_8_2_0_build_aarch64_src_cluster_websocket_events_bif
#define _usr_obj_ports_zeek_8_2_0_build_aarch64_src_cluster_websocket_events_bif
#include "zeek/zeek-bif.h"
#endif
#endif

namespace Cluster { extern zeek::EventHandlerPtr websocket_client_added;  }
namespace zeek::BifEvent::Cluster { void enqueue_websocket_client_added(zeek::analyzer::Analyzer* analyzer, zeek::IntrusivePtr<zeek::Val> endpoint, zeek::IntrusivePtr<zeek::Val> subscriptions);  }
namespace Cluster { extern zeek::EventHandlerPtr websocket_client_lost;  }
namespace zeek::BifEvent::Cluster { void enqueue_websocket_client_lost(zeek::analyzer::Analyzer* analyzer, zeek::IntrusivePtr<zeek::Val> endpoint, zeek_uint_t code, zeek::IntrusivePtr<zeek::StringVal> reason);  }

#endif
