// This file was automatically generated by bifcl from /usr/obj/ports/zeek-8.2.0/zeek-8.2.0/src/packet_analysis/protocol/teredo/events.bif (plugin mode).

#pragma once

#if defined(ZEEK_IN_NETVAR) || ! defined(_usr_obj_ports_zeek_8_2_0_build_aarch64_src_packet_analysis_protocol_teredo_events_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _usr_obj_ports_zeek_8_2_0_build_aarch64_src_packet_analysis_protocol_teredo_events_bif
#define _usr_obj_ports_zeek_8_2_0_build_aarch64_src_packet_analysis_protocol_teredo_events_bif
#include "zeek/zeek-bif.h"
#endif
#endif

extern zeek::EventHandlerPtr teredo_packet; 
namespace zeek::BifEvent { void enqueue_teredo_packet(zeek::analyzer::Analyzer* analyzer, zeek::Connection* outer, zeek::IntrusivePtr<zeek::Val> inner);  }
extern zeek::EventHandlerPtr new_teredo_state; 
namespace zeek::BifEvent { void enqueue_new_teredo_state(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c);  }
extern zeek::EventHandlerPtr teredo_authentication; 
namespace zeek::BifEvent { void enqueue_teredo_authentication(zeek::analyzer::Analyzer* analyzer, zeek::Connection* outer, zeek::IntrusivePtr<zeek::Val> inner);  }
extern zeek::EventHandlerPtr teredo_origin_indication; 
namespace zeek::BifEvent { void enqueue_teredo_origin_indication(zeek::analyzer::Analyzer* analyzer, zeek::Connection* outer, zeek::IntrusivePtr<zeek::Val> inner);  }
extern zeek::EventHandlerPtr teredo_bubble; 
namespace zeek::BifEvent { void enqueue_teredo_bubble(zeek::analyzer::Analyzer* analyzer, zeek::Connection* outer, zeek::IntrusivePtr<zeek::Val> inner);  }

#endif
