// See the file "COPYING" in the main distribution directory for copyright.

#pragma once

constexpr char ZEEK_SCRIPT_INSTALL_PATH[] = "/usr/local/share/zeek";
constexpr char ZEEK_PLUGIN_INSTALL_PATH[] = "/usr/local/lib/zeek/plugins";
constexpr char DEFAULT_ZEEKPATH[] = ".:/usr/local/share/zeek:/usr/local/share/zeek/policy:/usr/local/share/zeek/site:/usr/local/share/zeek/builtin-plugins";
constexpr char ZEEK_SPICY_MODULE_PATH[] = "";
constexpr char ZEEK_SPICY_LIBRARY_PATH[] = "";
constexpr char ZEEK_SPICY_DATA_PATH[] = "";
