// This file was automatically generated by bifcl from /usr/obj/ports/zeek-8.2.0/zeek-8.2.0/src/analyzer/protocol/dce-rpc/types.bif (plugin mode).

#pragma once

#if defined(ZEEK_IN_NETVAR) || ! defined(_usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_dce_rpc_types_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_dce_rpc_types_bif
#define _usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_dce_rpc_types_bif
#include "zeek/zeek-bif.h"
#endif
#endif

// NOLINTNEXTLINE(performance-enum-size)
namespace zeek::BifEnum::DCE_RPC { enum PType {
	REQUEST,
	PING,
	RESPONSE,
	FAULT,
	WORKING,
	NOCALL,
	REJECT,
	ACK,
	CL_CANCEL,
	FACK,
	CANCEL_ACK,
	BIND,
	BIND_ACK,
	BIND_NAK,
	ALTER_CONTEXT,
	ALTER_CONTEXT_RESP,
	AUTH3,
	SHUTDOWN,
	CO_CANCEL,
	ORPHANED,
	RTS,
}; }
namespace zeek::BifType::Enum::DCE_RPC { extern zeek::IntrusivePtr<zeek::EnumType> PType; }
// NOLINTNEXTLINE(performance-enum-size)
namespace zeek::BifEnum::DCE_RPC { enum IfID {
	unknown_if,
	epmapper,
	lsarpc,
	lsa_ds,
	mgmt,
	netlogon,
	samr,
	srvsvc,
	spoolss,
	drs,
	winspipe,
	wkssvc,
	oxid,
	ISCMActivator,
}; }
namespace zeek::BifType::Enum::DCE_RPC { extern zeek::IntrusivePtr<zeek::EnumType> IfID; }

#endif
