// This file was automatically generated by bifcl from /usr/obj/ports/zeek-8.2.0/zeek-8.2.0/src/packet_analysis/protocol/gtpv1/events.bif (plugin mode).

#pragma once

#if defined(ZEEK_IN_NETVAR) || ! defined(_usr_obj_ports_zeek_8_2_0_build_aarch64_src_packet_analysis_protocol_gtpv1_events_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _usr_obj_ports_zeek_8_2_0_build_aarch64_src_packet_analysis_protocol_gtpv1_events_bif
#define _usr_obj_ports_zeek_8_2_0_build_aarch64_src_packet_analysis_protocol_gtpv1_events_bif
#include "zeek/zeek-bif.h"
#endif
#endif

extern zeek::EventHandlerPtr new_gtpv1_state; 
namespace zeek::BifEvent { void enqueue_new_gtpv1_state(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c);  }
extern zeek::EventHandlerPtr gtpv1_message; 
namespace zeek::BifEvent { void enqueue_gtpv1_message(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> hdr);  }
extern zeek::EventHandlerPtr gtpv1_g_pdu_packet; 
namespace zeek::BifEvent { void enqueue_gtpv1_g_pdu_packet(zeek::analyzer::Analyzer* analyzer, zeek::Connection* outer, zeek::IntrusivePtr<zeek::Val> inner_gtp, zeek::IntrusivePtr<zeek::Val> inner_ip);  }
extern zeek::EventHandlerPtr gtpv1_create_pdp_ctx_request; 
namespace zeek::BifEvent { void enqueue_gtpv1_create_pdp_ctx_request(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> hdr, zeek::IntrusivePtr<zeek::Val> elements);  }
extern zeek::EventHandlerPtr gtpv1_create_pdp_ctx_response; 
namespace zeek::BifEvent { void enqueue_gtpv1_create_pdp_ctx_response(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> hdr, zeek::IntrusivePtr<zeek::Val> elements);  }
extern zeek::EventHandlerPtr gtpv1_update_pdp_ctx_request; 
namespace zeek::BifEvent { void enqueue_gtpv1_update_pdp_ctx_request(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> hdr, zeek::IntrusivePtr<zeek::Val> elements);  }
extern zeek::EventHandlerPtr gtpv1_update_pdp_ctx_response; 
namespace zeek::BifEvent { void enqueue_gtpv1_update_pdp_ctx_response(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> hdr, zeek::IntrusivePtr<zeek::Val> elements);  }
extern zeek::EventHandlerPtr gtpv1_delete_pdp_ctx_request; 
namespace zeek::BifEvent { void enqueue_gtpv1_delete_pdp_ctx_request(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> hdr, zeek::IntrusivePtr<zeek::Val> elements);  }
extern zeek::EventHandlerPtr gtpv1_delete_pdp_ctx_response; 
namespace zeek::BifEvent { void enqueue_gtpv1_delete_pdp_ctx_response(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> hdr, zeek::IntrusivePtr<zeek::Val> elements);  }

#endif
