// This file was automatically generated by bifcl from /usr/obj/ports/zeek-8.2.0/zeek-8.2.0/src/file_analysis/analyzer/extract/events.bif (plugin mode).

#pragma once

#if defined(ZEEK_IN_NETVAR) || ! defined(_usr_obj_ports_zeek_8_2_0_build_aarch64_src_file_analysis_analyzer_extract_events_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _usr_obj_ports_zeek_8_2_0_build_aarch64_src_file_analysis_analyzer_extract_events_bif
#define _usr_obj_ports_zeek_8_2_0_build_aarch64_src_file_analysis_analyzer_extract_events_bif
#include "zeek/zeek-bif.h"
#endif
#endif

extern zeek::EventHandlerPtr file_extraction_limit; 
namespace zeek::BifEvent { void enqueue_file_extraction_limit(zeek::analyzer::Analyzer* analyzer, zeek::IntrusivePtr<zeek::Val> f, zeek::IntrusivePtr<zeek::Val> args, zeek_uint_t limit, zeek_uint_t len);  }

#endif
