// This file was automatically generated by bifcl from /usr/obj/ports/zeek-8.2.0/zeek-8.2.0/src/analyzer/protocol/smb/smb1_com_transaction2_secondary.bif (plugin mode).

#pragma once

#if defined(ZEEK_IN_NETVAR) || ! defined(_usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_smb_smb1_com_transaction2_secondary_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_smb_smb1_com_transaction2_secondary_bif
#define _usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_smb_smb1_com_transaction2_secondary_bif
#include "zeek/zeek-bif.h"
#endif
#endif

extern zeek::EventHandlerPtr smb1_transaction2_secondary_request; 
namespace zeek::BifEvent { void enqueue_smb1_transaction2_secondary_request(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> hdr, zeek::IntrusivePtr<zeek::Val> args, zeek::IntrusivePtr<zeek::StringVal> parameters, zeek::IntrusivePtr<zeek::StringVal> data);  }
namespace zeek::BifType::Record::SMB1 { extern zeek::IntrusivePtr<zeek::RecordType> Trans2_Sec_Args; }

#endif
