// This file was automatically generated by bifcl from /usr/obj/ports/zeek-8.2.0/zeek-8.2.0/src/analyzer/protocol/mysql/events.bif (plugin mode).

#pragma once

#if defined(ZEEK_IN_NETVAR) || ! defined(_usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_mysql_events_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_mysql_events_bif
#define _usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_mysql_events_bif
#include "zeek/zeek-bif.h"
#endif
#endif

extern zeek::EventHandlerPtr mysql_command_request; 
namespace zeek::BifEvent { void enqueue_mysql_command_request(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek_uint_t command, zeek::IntrusivePtr<zeek::StringVal> arg);  }
extern zeek::EventHandlerPtr mysql_change_user; 
namespace zeek::BifEvent { void enqueue_mysql_change_user(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::StringVal> username);  }
extern zeek::EventHandlerPtr mysql_error; 
namespace zeek::BifEvent { void enqueue_mysql_error(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek_uint_t code, zeek::IntrusivePtr<zeek::StringVal> msg);  }
extern zeek::EventHandlerPtr mysql_ok; 
namespace zeek::BifEvent { void enqueue_mysql_ok(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek_uint_t affected_rows);  }
extern zeek::EventHandlerPtr mysql_eof; 
namespace zeek::BifEvent { void enqueue_mysql_eof(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, int is_intermediate);  }
extern zeek::EventHandlerPtr mysql_result_row; 
namespace zeek::BifEvent { void enqueue_mysql_result_row(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::Val> row);  }
extern zeek::EventHandlerPtr mysql_server_version; 
namespace zeek::BifEvent { void enqueue_mysql_server_version(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::StringVal> ver);  }
extern zeek::EventHandlerPtr mysql_handshake; 
namespace zeek::BifEvent { void enqueue_mysql_handshake(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::StringVal> username);  }
extern zeek::EventHandlerPtr mysql_ssl_request; 
namespace zeek::BifEvent { void enqueue_mysql_ssl_request(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c);  }
extern zeek::EventHandlerPtr mysql_auth_plugin; 
namespace zeek::BifEvent { void enqueue_mysql_auth_plugin(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, int is_orig, zeek::IntrusivePtr<zeek::StringVal> name, zeek::IntrusivePtr<zeek::StringVal> data);  }
extern zeek::EventHandlerPtr mysql_auth_switch_request; 
namespace zeek::BifEvent { void enqueue_mysql_auth_switch_request(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, zeek::IntrusivePtr<zeek::StringVal> name, zeek::IntrusivePtr<zeek::StringVal> data);  }
extern zeek::EventHandlerPtr mysql_auth_more_data; 
namespace zeek::BifEvent { void enqueue_mysql_auth_more_data(zeek::analyzer::Analyzer* analyzer, zeek::Connection* c, int is_orig, zeek::IntrusivePtr<zeek::StringVal> data);  }

#endif
