// This file was automatically generated by bifcl from /usr/obj/ports/zeek-8.2.0/zeek-8.2.0/src/broker/data.bif (alternative mode).

#pragma once

#if defined(ZEEK_IN_NETVAR) || ! defined(_usr_obj_ports_zeek_8_2_0_build_aarch64_src_broker_data_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _usr_obj_ports_zeek_8_2_0_build_aarch64_src_broker_data_bif
#define _usr_obj_ports_zeek_8_2_0_build_aarch64_src_broker_data_bif
#include "zeek/zeek-bif.h"
#endif
#endif

// NOLINTNEXTLINE(performance-enum-size)
namespace zeek::BifEnum::Broker { enum DataType {
	NONE,
	BOOL,
	INT,
	COUNT,
	DOUBLE,
	STRING,
	ADDR,
	SUBNET,
	PORT,
	TIME,
	INTERVAL,
	ENUM,
	SET,
	TABLE,
	VECTOR,
}; }
namespace zeek::BifType::Enum::Broker { extern zeek::IntrusivePtr<zeek::EnumType> DataType; }
namespace zeek::BifType::Record::Broker { extern zeek::IntrusivePtr<zeek::RecordType> Data; }
namespace zeek::BifType::Record::Broker { extern zeek::IntrusivePtr<zeek::RecordType> TableItem; }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __data_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __data_type_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __opaque_clone_through_serialization_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __set_create_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __set_clear_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __set_size_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __set_contains_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __set_insert_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __set_remove_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __set_iterator_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __set_iterator_last_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __set_iterator_next_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __set_iterator_value_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __table_create_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __table_clear_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __table_size_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __table_contains_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __table_insert_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __table_remove_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __table_lookup_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __table_iterator_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __table_iterator_last_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __table_iterator_next_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __table_iterator_value_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __vector_create_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __vector_clear_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __vector_size_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __vector_insert_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __vector_replace_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __vector_remove_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __vector_lookup_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __vector_iterator_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __vector_iterator_last_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __vector_iterator_next_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __vector_iterator_value_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __record_create_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __record_size_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __record_assign_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __record_lookup_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __record_iterator_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __record_iterator_last_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __record_iterator_next_bif(zeek::detail::Frame* frame, const zeek::Args*); }
namespace zeek::BifFunc::Broker { extern zeek::ValPtr __record_iterator_value_bif(zeek::detail::Frame* frame, const zeek::Args*); }

#endif
