// This file was automatically generated by bifcl from /usr/obj/ports/zeek-8.2.0/zeek-8.2.0/src/analyzer/protocol/snmp/types.bif (plugin mode).

#pragma once

#if defined(ZEEK_IN_NETVAR) || ! defined(_usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_snmp_types_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_snmp_types_bif
#define _usr_obj_ports_zeek_8_2_0_build_aarch64_src_analyzer_protocol_snmp_types_bif
#include "zeek/zeek-bif.h"
#endif
#endif

namespace zeek::BifType::Record::SNMP { extern zeek::IntrusivePtr<zeek::RecordType> Header; }
namespace zeek::BifType::Record::SNMP { extern zeek::IntrusivePtr<zeek::RecordType> HeaderV1; }
namespace zeek::BifType::Record::SNMP { extern zeek::IntrusivePtr<zeek::RecordType> HeaderV2; }
namespace zeek::BifType::Record::SNMP { extern zeek::IntrusivePtr<zeek::RecordType> HeaderV3; }
namespace zeek::BifType::Record::SNMP { extern zeek::IntrusivePtr<zeek::RecordType> UserSecurityParameters; }
namespace zeek::BifType::Record::SNMP { extern zeek::IntrusivePtr<zeek::RecordType> PDU; }
namespace zeek::BifType::Record::SNMP { extern zeek::IntrusivePtr<zeek::RecordType> TrapPDU; }
namespace zeek::BifType::Record::SNMP { extern zeek::IntrusivePtr<zeek::RecordType> BulkPDU; }
namespace zeek::BifType::Record::SNMP { extern zeek::IntrusivePtr<zeek::RecordType> ScopedPDU_Context; }
namespace zeek::BifType::Record::SNMP { extern zeek::IntrusivePtr<zeek::RecordType> ObjectValue; }
namespace zeek::BifType::Record::SNMP { extern zeek::IntrusivePtr<zeek::RecordType> Binding; }
namespace zeek::BifType::Vector::SNMP { extern zeek::IntrusivePtr<zeek::VectorType> Bindings; }

#endif
