// This file was automatically generated by bifcl from /usr/obj/ports/zeek-8.2.0/zeek-8.2.0/src/logging/writers/sqlite/sqlite.bif (plugin mode).

#pragma once

#if defined(ZEEK_IN_NETVAR) || ! defined(_usr_obj_ports_zeek_8_2_0_build_aarch64_src_logging_writers_sqlite_sqlite_bif)
#ifndef ZEEK_IN_NETVAR
#ifndef _usr_obj_ports_zeek_8_2_0_build_aarch64_src_logging_writers_sqlite_sqlite_bif
#define _usr_obj_ports_zeek_8_2_0_build_aarch64_src_logging_writers_sqlite_sqlite_bif
#include "zeek/zeek-bif.h"
#endif
#endif

namespace zeek::BifConst::LogSQLite { extern zeek::IntrusivePtr<zeek::StringVal> set_separator; }
namespace zeek::BifConst::LogSQLite { extern zeek::IntrusivePtr<zeek::StringVal> empty_field; }
namespace zeek::BifConst::LogSQLite { extern zeek::IntrusivePtr<zeek::StringVal> unset_field; }
// NOLINTNEXTLINE(performance-enum-size)
namespace zeek::BifEnum::LogSQLite { enum SQLiteSynchronous {
	SQLITE_SYNCHRONOUS_DEFAULT,
	SQLITE_SYNCHRONOUS_OFF,
	SQLITE_SYNCHRONOUS_NORMAL,
	SQLITE_SYNCHRONOUS_FULL,
	SQLITE_SYNCHRONOUS_EXTRA,
}; }
namespace zeek::BifType::Enum::LogSQLite { extern zeek::IntrusivePtr<zeek::EnumType> SQLiteSynchronous; }
// NOLINTNEXTLINE(performance-enum-size)
namespace zeek::BifEnum::LogSQLite { enum SQLiteJournalMode {
	SQLITE_JOURNAL_MODE_DEFAULT,
	SQLITE_JOURNAL_MODE_DELETE,
	SQLITE_JOURNAL_MODE_TRUNCATE,
	SQLITE_JOURNAL_MODE_PERSIST,
	SQLITE_JOURNAL_MODE_MEMORY,
	SQLITE_JOURNAL_MODE_WAL,
	SQLITE_JOURNAL_MODE_OFF,
}; }
namespace zeek::BifType::Enum::LogSQLite { extern zeek::IntrusivePtr<zeek::EnumType> SQLiteJournalMode; }
namespace zeek::BifConst::LogSQLite { extern zeek::IntrusivePtr<zeek::Val> synchronous; }
namespace zeek::BifConst::LogSQLite { extern zeek::IntrusivePtr<zeek::Val> journal_mode; }

#endif
